@@ -1,4 +1,4 @@
-/* $Id: micq.h,v 1.65.2.8 2003/09/19 22:42:34 kuhlmann Exp $ */
+/* $Id: micq.h,v 1.90 2005/01/11 21:57:57 kuhlmann Exp $ */
 
 #ifndef MICQ_H
 #define MICQ_H
@@ -22,19 +22,44 @@
     #include <sys/time.h>
   #endif
 #endif
+
 #include <stdio.h>
+
 #ifdef PREFER_PORTABLE_SNPRINTF
-  #include <snprintf.h>
+#include <snprintf.h>
 #endif
+
 #ifdef PREFER_PORTABLE_MEMMOVE
-  #include <memmove.h>
+#include <memmove.h>
 #endif
+
+#ifndef HAVE_TIMEGM
+#define HAVE_TIMEGM 1
+#define timegm portable_timegm
+time_t portable_timegm (struct tm *tm);
+#endif
+
+#ifndef HAVE_TIMELOCAL
+#define timelocal mktime
+#endif
+
+#ifndef HAVE_LOCALTIME_R
+#define HAVE_LOCALTIME_R 1
+#define localtime_r(t, s)      memcpy(s, localtime(t), sizeof(struct tm))
+#endif
+
+#ifndef _REENTRANT
+#define _REENTRANT 1
+#endif
+
 #ifndef __attribute__
 #if !defined(__GNUC__) || __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 7)
   #define __attribute__(x)
 #endif
 #endif
 
+#define ENABLE_DEBUG 1
+
 struct Queue_s;
 struct Event_s;
 struct Contact_s;
@@ -45,22 +70,51 @@ struct ConnectionPreferences_s;
 struct Cap_s;
 struct Extra_s;
 
-typedef struct Queue_s                 Queue;
-typedef struct Event_s                 Event;
-typedef struct Contact_s               Contact;
-typedef struct ContactGroup_s          ContactGroup;
-typedef struct Packet_s                Packet;
-typedef struct Connection_s            Connection;
-typedef struct Preferences_s           Preferences;
-typedef struct PreferencesConnection_s PreferencesConnection;
-typedef struct Cap_s                   Cap;
-typedef struct Extra_s                 Extra;
+typedef struct Queue_s        Queue;
+typedef struct Event_s        Event;
+typedef struct Opt_s          Opt;
+typedef struct Contact_s      Contact;
+typedef struct ContactGroup_s ContactGroup;
+typedef struct ContactAlias_s ContactAlias;
+typedef struct Packet_s       Packet;
+typedef struct Connection_s   Connection;
+typedef struct Preferences_s  Preferences;
+typedef struct Cap_s          Cap;
+typedef struct Extra_s        Extra;
 
 #include "datatype.h"
 #include <stdlib.h>
-#include "mreadline.h"
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+
+#if ENABLE_DEBUG && HAVE_VARIADIC_MACRO
+#define DEBUGARGS    , __FILE__, __LINE__
+#define DEBUGPARAM   , const char *debugfile, int debugline
+#define DEBUG0ARGS   __FILE__, __LINE__
+#define DEBUG0PARAM  const char *debugfile, int debugline
+#define DEBUGFOR     , debugfile, debugline
+#define DEBUGNONE    , "", 0
+#define Debug(l,f,...) DebugReal (l, f " {%s:%d}<<{%s:%d}" , ## __VA_ARGS__, \
+                                  __FILE__, __LINE__, debugfile, debugline)
+#define DebugH(l,f,...) DebugReal (l, f " {%s:%d}<|" , ## __VA_ARGS__, \
+                                   __FILE__, __LINE__)
+#else
+#define DEBUGARGS
+#define DEBUGPARAM
+#define DEBUG0ARGS
+#define DEBUG0PARAM  void
+#define DEBUGFOR
+#define DEBUGNONE
+#define DebugH DebugReal
+#define Debug DebugReal
+#endif
+
 #include "msg_queue.h"
-#include "mselect.h"
+#include "mreadline.h"
 #include "color.h"
 
 #define SOUND_ON 1
@@ -101,8 +155,6 @@ typedef struct Extra_s                 Extra;
 #define STATUS_FFC         STATUSF_FFC
 #define STATUS_ONLINE     0x00000000
 
-#define NOW -1
-
 #define MSGF_MASS         0x8000
 #define MSGF_GETAUTO      0x0300
 
@@ -124,6 +176,7 @@ typedef struct Extra_s                 Extra;
 #define MSG_GET_NA        0xea
 #define MSG_GET_DND       0xeb
 #define MSG_GET_FFC       0xec
+/* mICQ extension */
 #define MSG_GET_PEEK      0xf0
 #define MSG_GET_VER       0xf4
 /* Licq extension */
@@ -134,6 +187,9 @@ typedef struct Extra_s                 Extra;
 #define VIS_LIST_UPDATE         0x02
 #define CONT_LIST_UPDATE        0x00
 
+#define LOG_MAX_PATH 255
+#define DSCSIZ 192 /* Maximum length of log file descriptor lines. */
+
 /*
  * Should you decide to move a global variable from one struct to another,
  * change it manually in the header file, and then in the src directory
@@ -145,20 +201,20 @@ typedef struct Extra_s                 Extra;
 /* user interface global state variables */
 typedef struct
 {
-        UDWORD last_rcvd_uin;
-        UDWORD last_sent_uin;
-        char*  last_message_sent;
-        UDWORD last_message_sent_type;
-        time_t start_time;
-        UDWORD away_time_prev;
-        UBYTE  reconnect_count;
-        BOOL   quit;
-        UDWORD packets;
-        time_t idle_val;
-        UDWORD idle_msgs;
-        UBYTE  idle_flag;
-        char * idle_uins;
-        int    nick_len;   /* this *must* be an int */
+    Contact *last_rcvd;
+    Contact *last_sent;
+    char    *last_message_sent;
+    UDWORD   last_message_sent_type;
+    time_t   start_time;
+    UDWORD   away_time_prev;
+    UBYTE    reconnect_count;
+    BOOL     quit;
+    UDWORD   packets, events;
+    time_t   idle_val;
+    UDWORD   idle_msgs;
+    UBYTE    idle_flag;
+    char    *idle_uins;
+    int      nick_len;   /* this *must* be an int */
 } user_interface_state;
 
 extern user_interface_state uiG;
@@ -184,3 +240,4 @@ enum logtype
 #endif /* MICQ_H */
 
 #include "i18n.h"
+#include "util_str.h"
