@@ -2,7 +2,7 @@
 /*
  * Assemble outgoing and dissect incoming packets.
  *
- * mICQ Copyright (C) © 2001,2002,2003 Rüdiger Kuhlmann
+ * mICQ Copyright (C) © 2001-2005 Rüdiger Kuhlmann
  *
  * mICQ is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -13,12 +13,21 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  * License for more details.
  *
+ * In addition, as a special exception permission is granted to link the
+ * code of this release of mICQ with the OpenSSL project's "OpenSSL"
+ * library, and distribute the linked executables.  You must obey the GNU
+ * General Public License in all respects for all of the code used other
+ * than "OpenSSL".  If you modify this file, you may extend this exception
+ * to your version of the file, but you are not obligated to do so.  If you
+ * do not wish to do so, delete this exception statement from your version
+ * of this file.
+ *
  * You should have received a copy of the GNU General Public License
  * along with this package; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  * 02111-1307, USA.
  *
- * $Id: packet.c,v 1.39.2.4 2003/05/12 14:38:39 kuhlmann Exp $
+ * $Id: packet.c,v 1.60 2005/02/11 23:33:11 kuhlmann Exp $
  *
  * Note: alle read strings need to be free()ed.
  */
@@ -27,11 +36,9 @@
 #include "packet.h"
 #include "preferences.h"
 #include "util_ui.h"
-#include "util_str.h"
+#include "contact.h"
 #include "buildmark.h"
 #include <assert.h>
-#include <stdio.h>
-#include <string.h>
 
 #define cap_id   "\x82\x22\x44\x45\x53\x54\x00\x00"
 #define cap_none "\x00\x00\x00\x00\x00\x00\x00\x00"
@@ -43,35 +50,74 @@
 
 static Cap caps[CAP_MAX] =
 {
-    { CAP_NONE,        16, cap_none cap_none,                   "CAP_NONE"        },
-    { CAP_AIM_VOICE,   16, cap_aim "\x41" cap_mid cap_id,       "CAP_AIM_VOICE"   },
-    { CAP_AIM_SFILE,   16, cap_aim "\x43" cap_mid cap_id,       "CAP_AIM_SFILE"   },
-    { CAP_ISICQ,       16, cap_aim "\x44" cap_mid cap_id,       "CAP_ISICQ"       },
-    { CAP_AIM_IMIMAGE, 16, cap_aim "\x45" cap_mid cap_id,       "CAP_AIM_IMIMAGE" },
-    { CAP_AIM_BUDICON, 16, cap_aim "\x46" cap_mid cap_id,       "CAP_AIM_BUDICON" },
-    { CAP_AIM_STOCKS,  16, cap_aim "\x47" cap_mid cap_id,       "CAP_AIM_STOCKS"  },
-    { CAP_AIM_GETFILE, 16, cap_aim "\x48" cap_mid cap_id,       "CAP_AIM_GETFILE" },
-    { CAP_SRVRELAY,    16, cap_aim "\x49" cap_mid cap_id,       "CAP_SRVRELAY"    },
-    { CAP_AIM_GAMES,   16, cap_aim "\x4a" cap_mid cap_id,       "CAP_AIM_GAMES"   },
-    { CAP_AIM_SBUD,    16, cap_aim "\x4b" cap_mid cap_id,       "CAP_AIM_SBUD"    },
-    { CAP_UTF8,        16, cap_aim "\x4e" cap_mid cap_id,       "CAP_UTF8"        },
-    { CAP_RTFMSGS,     16, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x92", "CAP_RTFMSGS"     },
-    { CAP_IS_2001,     16, "\x2e\x7a\x64\x75\xfa\xdf\x4d\xc8\x88\x6f\xea\x35\x95\xfd\xb6\xdf", "CAP_IS_2001"     },
-    { CAP_STR_2001,    16, "\xa0\xe9\x3f\x37" cap_mstr cap_str, "CAP_STR_2001"    },
-    { CAP_STR_2002,    16, "\x10\xcf\x40\xd1" cap_mstr cap_str, "CAP_STR_2002"    },
-    { CAP_AIM_CHAT,    16, "\x74\x8f\x24\x20\x62\x87\x11\xd1" cap_id,                          "CAP_AIM_CHAT"    },
-    { CAP_IS_WEB,      16, "\x56\x3f\xc8\x09\x0b\x6f\x41\xbd\x9f\x79\x42\x26\x09\xdf\xa2\xf3", "CAP_IS_WEB"      },
-    { CAP_TRILL_CRYPT, 16, "\xf2\xe7\xc7\xf4\xfe\xad\x4d\xfb\xb2\x35\x36\x79\x8b\xdf\x00\x00", "CAP_TRILL_CRYPT" },
-    { CAP_TRILL_2,     16, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x09", "CAP_TRILL_2"     },
-    { CAP_LICQ,        16, "\x09\x49\x13\x49" cap_mid cap_id,   "CAP_LICQ"        },
-    { CAP_SIM,         15, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x48", "CAP_SIM"         },
-    { CAP_MACICQ,      16, "\xdd\x16\xf2\x02\x84\xe6\x11\xd4\x90\xdb\x00\x10\x4b\x9b\x4b\x7d", "CAP_MACICQ"      },
-    { CAP_MICQ,        12, "mICQ \xa9 R.K. \x00\x00\x00\x00",   "CAP_MICQ"        },
-    { CAP_KXICQ,       16, "\x09\x49\x13\x44" cap_mid cap_id,   "CAP_KXICQ"},
-    { 0, 0, NULL, NULL }
+    { CAP_NONE,        16, cap_none cap_none,                   "CAP_NONE",        NULL },
+    { CAP_AIM_VOICE,   16, cap_aim "\x41" cap_mid cap_id,       "CAP_AIM_VOICE",   NULL },
+    { CAP_AIM_SFILE,   16, cap_aim "\x43" cap_mid cap_id,       "CAP_AIM_SFILE",   NULL },
+    { CAP_ISICQ,       16, cap_aim "\x44" cap_mid cap_id,       "CAP_ISICQ",       NULL },
+    { CAP_AIM_IMIMAGE, 16, cap_aim "\x45" cap_mid cap_id,       "CAP_AIM_IMIMAGE", NULL },
+    { CAP_AIM_BUDICON, 16, cap_aim "\x46" cap_mid cap_id,       "CAP_AIM_BUDICON", NULL },
+    { CAP_AIM_STOCKS,  16, cap_aim "\x47" cap_mid cap_id,       "CAP_AIM_STOCKS",  NULL },
+    { CAP_AIM_GETFILE, 16, cap_aim "\x48" cap_mid cap_id,       "CAP_AIM_GETFILE", NULL },
+    { CAP_SRVRELAY,    16, cap_aim "\x49" cap_mid cap_id,       "CAP_SRVRELAY",    NULL },
+    { CAP_AIM_GAMES,   16, cap_aim "\x4a" cap_mid cap_id,       "CAP_AIM_GAMES",   NULL },
+    { CAP_AIM_SBUD,    16, cap_aim "\x4b" cap_mid cap_id,       "CAP_AIM_SBUD",    NULL },
+    { CAP_UTF8,        16, cap_aim "\x4e" cap_mid cap_id,       "CAP_UTF8",        NULL },
+    { CAP_RTFMSGS,     16, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x92", "CAP_RTFMSGS",     NULL },
+    { CAP_IS_2001,     16, "\x2e\x7a\x64\x75\xfa\xdf\x4d\xc8\x88\x6f\xea\x35\x95\xfd\xb6\xdf", "CAP_IS_2001",     NULL },
+    { CAP_STR_2001,    16, "\xa0\xe9\x3f\x37" cap_mstr cap_str, "CAP_STR_2001",    NULL },
+    { CAP_STR_2002,    16, "\x10\xcf\x40\xd1" cap_mstr cap_str, "CAP_STR_2002",    NULL },
+    { CAP_AIM_CHAT,    16, "\x74\x8f\x24\x20\x62\x87\x11\xd1" cap_id,                          "CAP_AIM_CHAT",    NULL },
+    { CAP_TYPING,      16, "\x56\x3f\xc8\x09\x0b\x6f\x41\xbd\x9f\x79\x42\x26\x09\xdf\xa2\xf3", "CAP_TYPING",      NULL },
+    { CAP_TRILL_CRYPT, 16, "\xf2\xe7\xc7\xf4\xfe\xad\x4d\xfb\xb2\x35\x36\x79\x8b\xdf\x00\x00", "CAP_TRILL_CRYPT", NULL },
+    { CAP_TRILL_2,     16, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x09", "CAP_TRILL_2",     NULL },
+    { CAP_LICQ,        16, "\x09\x49\x13\x49" cap_mid cap_id,   "CAP_LICQ",        NULL },
+    { CAP_LICQNEW,     12, "Licq client \x00\x00\x00\x00",      "CAP_LICQNEW",     NULL },
+    { CAP_SIM,         15, "\x97\xb1\x27\x51\x24\x3c\x43\x34\xad\x22\xd6\xab\xf7\x3f\x14\x48", "CAP_SIM",         NULL },
+    { CAP_SIMNEW,      12, "SIM client  \x00\x00\x00\x00",      "CAP_SIMNEW",      NULL },
+    { CAP_MACICQ,      16, "\xdd\x16\xf2\x02\x84\xe6\x11\xd4\x90\xdb\x00\x10\x4b\x9b\x4b\x7d", "CAP_MACICQ",      NULL },
+    { CAP_MICQ,        12, "mICQ \xa9 R.K. \x00\x00\x00\x00",   "CAP_MICQ",        NULL },
+    { CAP_KXICQ,       16, "\x09\x49\x13\x44" cap_mid cap_id,   "CAP_KXICQ",       NULL },
+    { CAP_KOPETE,      12, "Kopete ICQ  \x00\x00\x00\x00",      "CAP_KOPETE",      NULL },
+    { CAP_IMSECURE,    12, "IMsecureCphr\x00\x00\x00\x00",      "CAP_IMSECURE",    NULL },
+    { CAP_ARQ,          9, "&RQinside\x00\x00\x00\x00\x00\x00\x00", "CAP_ARQ",     NULL },
+    { CAP_UTF8ii,      16, cap_aim "\x4e" cap_mid "\x82\x22\x44\x45\x53\x54ii", "CAP_UTF8ii", NULL },
+    { CAP_AIM_INTER,   16, cap_aim "\x4d" cap_mid cap_id,       "CAP_AIM_INTER",   NULL },
+    { CAP_WIERD1,      16, "\x17\x8c\x2d\x9b\xda\xa5\x45\xbb\x8d\xdb\xf3\xbd\xbd\x53\xa1\x0a", "CAP_WIERD1", NULL },
+    { CAP_WIERD2,      16, "\x1a\x09\x3c\x6c\xd7\xfd\x4e\xc5\x9d\x51\xa6\x47\x4e\x34\xf5\xa0", "CAP_WIERD2", NULL },
+    { CAP_WIERD3,      16, "\x67\x36\x15\x15\x61\x2d\x4c\x07\x8f\x3d\xbd\xe6\x40\x8e\xa0\x41", "CAP_WIERD3", NULL },
+    { CAP_WIERD4,      16, "\xe3\x62\xc1\xe9\x12\x1a\x4b\x94\xa6\x26\x7a\x74\xde\x24\x27\x0d", "CAP_WIERD4", NULL },
+    { CAP_WIERD5,      16, "\xb9\x97\x08\xb5\x3a\x92\x42\x02\xb0\x69\xf1\xe7\x57\xbb\x2e\x17", "CAP_WIERD5", NULL },
+    { CAP_11,          16, "\x01\x01\x01\x01\x01\x01\x19\x04\x4a\x16\xed\x79\x2c\xb1\x71\x01", "CAP_11", NULL },
+    { CAP_12,          16, "\x02\x02\x02\x02\x02\x02\xb3\xf8\x53\x44\x7f\x0d\x2d\x83\xbd\x76", "CAP_12", NULL },  
+    
+    
+    /* CAP_UNK_37: 01 01 01 01 01 01 19 04  4a 16 ed 79 2c b1 71 01  "........ J.íy,±q." IMsecure?
+    // CAP_UNK_38: 02 02 02 02 02 02 b3 f8  53 44 7f 0d 2d 83 bd 76  "......³ø SD..-.½v" IMsecure?
+    // CAP_UNK_39: b6 07 43 78 f5 0c 4a c7  90 92 59 38 50 2d 05 91  "¶.Cxõ.JÇ ..Y8P-.." ICQ Lite
+    // CAP_UNK_40: 80 53 7d e2 a4 67 4a 76  b3 54 6d fd 07 5f 5e c6  ".S}â¤gJv ³Tmý._^Æ" ICQ Lite
+    */
+
+    { 0, 0, NULL, NULL, NULL }
 };
 
-Packet *PacketC (void)
+static str_s packetstr[] =
+{
+    { NULL, 0, 0 },
+    { NULL, 0, 0 },
+    { NULL, 0, 0 },
+    { NULL, 0, 0 },
+    { "<invalidlen>", 12, 0 },
+    { "", 0, 0 },
+};
+
+static int packetstrind = 0;
+
+#define PACKETMAXSTR 4
+#define PACKETSTRINVALID 4
+#define PACKETSTREMPTY 5
+
+#undef PacketC
+Packet *PacketC (DEBUG0PARAM)
 {
     Packet *pak;
     
@@ -84,15 +130,16 @@ Packet *PacketC (void)
     return pak;
 }
 
-Packet *PacketCreate (const void *data, UDWORD len)
+#undef PacketCreate
+Packet *PacketCreate (str_t str DEBUGPARAM)
 {
     Packet *newpak;
     
     newpak = calloc (1, sizeof (Packet));
     assert (newpak);
     
-    memcpy (newpak->data, data, len);
-    newpak->len = len;
+    memcpy (newpak->data, str->txt, str->len);
+    newpak->len = str->len;
     
     Debug (DEB_PACKET, "<-+- %p create", newpak);
     uiG.packets++;
@@ -100,14 +147,16 @@ Packet *PacketCreate (const void *data, UDWORD len)
     return newpak;
 }
 
-void PacketD (Packet *pak)
+#undef PacketD
+void PacketD (Packet *pak DEBUGPARAM)
 {
     Debug (DEB_PACKET, "---> %p free", pak);
     uiG.packets--;
     free (pak);
 }
 
-Packet *PacketClone (const Packet *pak)
+#undef PacketClone
+Packet *PacketClone (const Packet *pak DEBUGPARAM)
 {
     Packet *newpak;
     
@@ -213,7 +262,7 @@ void PacketWriteCap (Packet *pak, Cap *cap)
     assert (pak);
     assert (cap);
 
-    PacketWriteData (pak, (const char *)(cap->var ? cap->var : cap->cap), 16);
+    PacketWriteData (pak, (cap->var ? (const char *)cap->var : cap->cap), 16);
 }
 
 void PacketWriteData (Packet *pak, const char *data, UWORD len)
@@ -269,11 +318,11 @@ void PacketWriteLLNTS (Packet *pak, const char *data)
     PacketWrite1 (pak, 0);
 }
 
-void PacketWriteUIN (Packet *pak, UDWORD uin)
+void PacketWriteCont (Packet *pak, Contact *cont)
 {
     const char *str;
     
-    str = s_sprintf ("%ld", uin);
+    str = s_sprintf ("%ld", cont->uin);
     PacketWrite1 (pak, strlen (str));
     PacketWriteData (pak, str, strlen (str));
 }
@@ -315,6 +364,15 @@ void PacketWriteLenDone (Packet *pak)
     pak->tpos = pos;
 }
 
+void PacketWriteBLenDone (Packet *pak)
+{
+    UWORD pos;
+    
+    pos = PacketReadAtB2 (pak, pak->tpos);
+    PacketWriteAtB2 (pak, pak->tpos, pak->wpos - pak->tpos - 2);
+    pak->tpos = pos;
+}
+
 void PacketWriteLen4 (Packet *pak)
 {
     UWORD pos;
@@ -516,105 +574,163 @@ Cap *PacketReadCap (Packet *pak)
 #endif
         }
         else
-            if (id == CAP_MAX)
-                return &caps[0];
-            else
-                break;
+            break;
+    if (id == CAP_MAX)
+        return &caps[0];
     
     p = malloc (16);
     assert (p);
     memcpy (p, cap, 16);
 
     caps[id].id = id;
-    caps[id].cap = (const UBYTE *)p;
+    caps[id].cap = p;
     caps[id].len = 16;
     caps[id].name = strdup (s_sprintf ("CAP_UNK_%d", id));
     return &caps[id];
 }
 
-void PacketReadData (Packet *pak, char *buf, UWORD len)
+void PacketReadData (Packet *pak, str_t str, UWORD len)
 {
     assert (pak);
     
     if (pak->rpos + len > PacketMaxData)
+    {
+        if (str)
+        {
+            str->len = 0;
+            if (str->max)
+                *str->txt = '\0';
+        }
         return;
-    
-    if (buf)
-        memcpy (buf, pak->data + pak->rpos, len);
-
+    }
+    if (str)
+    {
+        s_init (str, "", len + 1);
+        if (str->max > len)
+        {
+            str->len = len;
+            str->txt[len] = '\0';
+            memcpy (str->txt, pak->data + pak->rpos, len);
+        }
+    }
     pak->rpos += len;
 }
 
-char *PacketReadStrB (Packet *pak)
+strc_t PacketReadB2Str (Packet *pak, str_t str)
 {
-    UWORD len;
-    char *str;
+    unsigned int len;
     
     len = PacketReadB2 (pak);
-
     if (pak->rpos + len >= PacketMaxData)
-        return strdup ("<invalidlen>");
+    {
+        if (!str)
+            return &packetstr[PACKETSTRINVALID];
+        s_init (str, packetstr[PACKETSTRINVALID].txt, 0);
+        return str;
+    }
 
-    str = malloc (len + 1);
-    assert (str);
+    if (!str)
+    {
+        packetstrind %= 4;
+        str = &packetstr[packetstrind++];
+    }
+
+    s_init (str, "", len + 2);
+    assert (str->max >= len + 2);
     
     PacketReadData (pak, str, len);
-    str[len] = '\0';
+    str->txt[str->len = len] = '\0';
     
     return str;
 }
 
-char *PacketReadLNTS (Packet *pak)
+strc_t PacketReadL2Str (Packet *pak, str_t str)
 {
-    UWORD len;
-    char *str;
+    unsigned int len;
     
     len = PacketRead2 (pak);
-
     if (pak->rpos + len >= PacketMaxData)
-        str = strdup ("<invalidlen>");
+    {
+        if (!str)
+            return &packetstr[PACKETSTRINVALID];
+        s_init (str, packetstr[PACKETSTRINVALID].txt, 0);
+        return str;
+    }
     if (!len)
-        return strdup ("");
+    {
+        if (!str)
+            return &packetstr[PACKETSTREMPTY];
+        s_init (str, "", 0);
+        return str;
+    }
+    
+    if (!str)
+    {
+        packetstrind %= 4;
+        str = &packetstr[packetstrind++];
+    }
 
-    str = malloc (len);
-    assert (str);
+    s_init (str, "", len + 2);
+    assert (str->max >= len + 2);
 
     PacketReadData (pak, str, len);
-    str[len - 1] = '\0';
+    str->txt[str->len = len - 1] = '\0';
 
     return str;
 }
 
-char *PacketReadDLStr (Packet *pak)
+strc_t PacketReadL4Str (Packet *pak, str_t str)
 {
-    UWORD len;
-    char *str;
+    size_t len;
     
     len = PacketRead4 (pak);
-
     if (pak->rpos + len >= PacketMaxData)
-        return strdup ("<invalidlen>");
+    {
+        if (!str)
+            return &packetstr[PACKETSTRINVALID];
+        s_init (str, packetstr[PACKETSTRINVALID].txt, 0);
+        return str;
+    }
 
-    str = malloc (len + 1);
-    assert (str);
+    if (!str)
+    {
+        packetstrind %= 4;
+        str = &packetstr[packetstrind++];
+    }
+
+    s_init (str, "", len + 2);
+    assert (str->max >= len + 2);
     
     PacketReadData (pak, str, len);
-    str[len] = '\0';
+    str->txt[str->len = len] = '\0';
 
     return str;
 }
 
-UDWORD PacketReadUIN (Packet *pak)
+strc_t PacketReadUIN (Packet *pak)
 {
     UBYTE len = PacketRead1 (pak);
-    char *str = malloc (len + 1);
-    UDWORD uin;
+    str_t str;
 
+    packetstrind %= 4;
+    str = &packetstr[packetstrind++];
+    s_init (str, "", len + 2);
     PacketReadData (pak, str, len);
-    str[len] = '\0';
-    uin = atoi (str);
-    free (str);
-    return uin;
+    str->txt[len] = '\0';
+    return str;
+}
+
+Contact *PacketReadCont (Packet *pak, Connection *serv)
+{
+    UBYTE len = PacketRead1 (pak);
+    Contact *cont;
+    str_s str = { NULL, 0, 0 };
+
+    PacketReadData (pak, &str, len);
+    str.txt[len] = '\0';
+    cont = ContactScreen (serv, str.txt);
+    s_done (&str);
+    return cont;
 }
 
 UWORD PacketReadPos (const Packet *pak)
@@ -692,52 +808,59 @@ UDWORD PacketReadAtB4 (const Packet *pak, UWORD at)
     return data;
 }
 
-void PacketReadAtData (const Packet *pak, UWORD at, char *buf, UWORD len)
+void PacketReadAtData (const Packet *pak, UWORD at, str_t str, UWORD len)
 {
     assert (pak);
 
     if (at + len > PacketMaxData)
+    {
+        if (str)
+        {
+            str->len = 0;
+            if (str->max)
+                *str->txt = '\0';
+        }
         return;
-
-    memcpy (buf, pak->data + at, len);
+    }
+    if (str)
+    {
+        s_init (str, "", len + 1);
+        if (str->max > len)
+            memcpy (str->txt, pak->data + at, len);
+    }
 }
 
-char *PacketReadAtStrB (const Packet *pak, UWORD at)
+strc_t PacketReadAtL2Str (const Packet *pak, UWORD at, str_t str)
 {
-    UWORD len;
-    char *str;
-    
-    len = PacketReadAtB2 (pak, at);
-
-    if (!len)
-        return "";
-    
-    str = malloc (len + 1);
-    assert (str);
-    
-    PacketReadAtData (pak, at + 2, str, len);
-    str[len] = '\0';
-
-    return str;
-}
-
-char *PacketReadAtLNTS (Packet *pak, UWORD at)
-{
-    UWORD len;
-    char *str;
+    unsigned int len;
     
     len = PacketReadAt2 (pak, at);
-
     if (at + 2 + len >= PacketMaxData)
-        return strdup ("<invalid>");
+    {
+        if (!str)
+            return &packetstr[PACKETSTRINVALID];
+        s_init (str, packetstr[PACKETSTRINVALID].txt, 0);
+        return str;
+    }
     if (!len)
-        return strdup ("");
+    {
+        if (!str)
+            return &packetstr[PACKETSTREMPTY];
+        s_init (str, "", 0);
+        return str;
+    }
 
-    str = malloc (len);
-    assert (str);
+    if (!str)
+    {
+        packetstrind %= 4;
+        str = &packetstr[packetstrind++];
+    }
+
+    s_init (str, "", len + 2);
+    assert (str->max >= len + 2);
 
     PacketReadAtData (pak, at + 2, str, len);
-    str[len - 1] = '\0';
+    str->txt[str->len = len - 1] = '\0';
 
     return str;
 }
