@@ -12,22 +12,28 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  * License for more details.
  *
+ * In addition, as a special exception permission is granted to link the
+ * code of this release of mICQ with the OpenSSL project's "OpenSSL"
+ * library, and distribute the linked executables.  You must obey the GNU
+ * General Public License in all respects for all of the code used other
+ * than "OpenSSL".  If you modify this file, you may extend this exception
+ * to your version of the file, but you are not obligated to do so.  If you
+ * do not wish to do so, delete this exception statement from your version
+ * of this file.
+ *
  * You should have received a copy of the GNU General Public License
  * along with this package; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  * 02111-1307, USA.
  *
- * $Id: peer_file.c,v 1.38.2.6 2003/09/19 19:40:34 kuhlmann Exp $
+ * $Id: peer_file.c,v 1.69 2005/02/28 19:06:39 kuhlmann Exp $
  */
 
 #include "micq.h"
 
 #ifdef ENABLE_PEER2PEER
 
-#include <stdio.h>
 #include <unistd.h>
-#include <string.h>
-#include <time.h>
 #include <assert.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -45,12 +51,11 @@
 #endif
 
 #include "preferences.h"
-#include "session.h"
+#include "connection.h"
 #include "packet.h"
 #include "util_ui.h"
 #include "util_io.h"
-#include "util_str.h"
-#include "util_extra.h"
+#include "util_rl.h"
 #include "util.h"
 #include "icq_response.h"
 #include "tcp.h"
@@ -59,7 +64,11 @@
 static void PeerFileDispatchClose   (Connection *ffile);
 static void PeerFileDispatchDClose  (Connection *ffile);
 static void PeerFileIODispatchClose (Connection *ffile);
+static void PeerFileDispatchW       (Connection *fpeer);
+static BOOL PeerFileError (Connection *fpeer, UDWORD rc, UDWORD flags);
 
+#define FAIL(x) { err = x; break; }
+#define PeerFileClose TCPClose
 
 /*
  * Create a new file listener unless one already exists.
@@ -70,79 +79,115 @@ Connection *PeerFileCreate (Connection *serv)
     
     ASSERT_ANY_SERVER(serv);
     
-    if (!serv->assoc || serv->assoc->ver < 6)
+    if (!serv->assoc || serv->assoc->version < 6)
         return NULL;
     
     if ((flist = ConnectionFind (TYPE_FILELISTEN, 0, serv)))
         return flist;
     
-    flist = ConnectionClone (serv->assoc, TYPE_FILELISTEN);
+    flist = ConnectionClone (serv, TYPE_FILELISTEN);
     if (!flist)
         return NULL;
 
     if (prG->verbose)
-        M_printf (i18n (2082, "Opening file listener connection at localhost:%ld... "), serv->assoc->spref->port);
+        rl_printf (i18n (2519, "Opening file listener connection at %slocalhost%s:%s%ld%s... "),
+                  COLQUOTE, COLNONE, COLQUOTE, serv->assoc->pref_port, COLNONE);
 
-    flist->spref       = NULL;
-    flist->parent      = serv;
-    flist->connect     = 0;
-    flist->flags       = 0;
-    flist->our_seq     = -1;
-    flist->our_session = 0;
-    flist->port        = serv->assoc->spref->port;
-    s_repl (&flist->server, NULL);
-    flist->ip          = 0;
-    flist->dispatch    = &TCPDispatchMain;
-    flist->reconnect   = &TCPDispatchReconn;
-    flist->close       = &PeerFileDispatchClose;
+    flist->our_seq  = -1;
+    flist->version  = serv->assoc->version;
+    flist->cont     = serv->assoc->cont;
+    flist->port     = serv->assoc->pref_port;
+    flist->dispatch = &TCPDispatchMain;
+    flist->close    = &PeerFileDispatchClose;
     
     UtilIOConnectTCP (flist);
     
     return flist;
 }
 
+/*
+ * Handles a timeout.
+ */
+
+void PeerFileTO (Event *event)
+{
+    QueueEnqueue (event);
+    QueueRelease (event);
+    event->callback = NULL;
+}
+
+/*
+ * Handles user reaction to incoming file request
+ */
+void PeerFileUser (UDWORD seq, Contact *cont, const char *reason, Connection *serv)
+{
+    Event *event;
+
+    if (!(event = QueueDequeue2 (serv, QUEUE_USERFILEACK, seq, cont)))
+    {
+        rl_printf (i18n (2258, "No pending incoming file transfer request for %s with (sequence %ld) found.\n"),
+                  cont ? cont->nick : "<?>", seq);
+    }
+    else
+    {
+        if (reason)
+            event->opt = OptSetVals (event->opt, CO_FILEACCEPT, 0, CO_REFUSE, reason, 0);
+        else
+            event->opt = OptSetVals (event->opt, CO_FILEACCEPT, 1, 0);
+        QueueEnqueue (event);
+        QueueRelease (event);
+    }
+}
+
 /*
  * Handles an incoming file request.
  */
 UBYTE PeerFileIncAccept (Connection *list, Event *event)
 {
     Connection *flist, *fpeer, *serv;
-    Extra *extra;
+    Opt *opt;
     Contact *cont;
-    UDWORD bytes = ExtraGet (event->extra, EXTRA_FILETRANS);
-    const char *files = ExtraGetS (event->extra, EXTRA_FILETRANS);
-    const char *txt;
+    UDWORD opt_bytes, opt_acc;
+    const char *opt_files;
+
+    if (!OptGetVal (event->opt, CO_BYTES, &opt_bytes))
+        opt_bytes = 0;
+    if (!OptGetStr (event->opt, CO_MSGTEXT, &opt_files))
+        opt_files = "";
 
     ASSERT_MSGLISTEN(list);
     
     serv  = list->parent;
-    cont  = ContactUIN (serv, event->uin);
+    cont  = event->cont;
     flist = PeerFileCreate (serv);
 
-    extra = ExtraSet (NULL, 0, bytes, files);
-
-    if (((txt = ExtraGetS (event->extra, EXTRA_FILEACCEPT)) && *txt) || !cont || !flist
-        || !ExtraGet (event->extra, EXTRA_FILEACCEPT)
+    if ((event->wait && !OptGetVal (event->wait->opt, CO_FILEACCEPT, &opt_acc))
+        || !opt_acc || !cont || !flist
         || !(fpeer = ConnectionClone (flist, TYPE_FILEDIRECT)))
     {
-        if (!txt || !*txt)
-            event->extra = ExtraSet (event->extra, EXTRA_FILEACCEPT, 0, txt = "auto-refused");
-        IMIntMsg (cont, serv, NOW, STATUS_OFFLINE, INT_FILE_REJING, txt, extra);
+        const char *txt;
+        opt = OptSetVals (NULL, CO_MSGTEXT, opt_files, 0);
+        if (!OptGetStr (event->wait->opt, CO_REFUSE, &txt))
+            txt = "";
+        IMIntMsg (cont, serv, NOW, STATUS_OFFLINE, INT_FILE_REJING, txt, opt);
         return FALSE;
     }
-    
     ASSERT_FILELISTEN (flist);
     ASSERT_FILEDIRECT (fpeer);
     
-    fpeer->port    = 0;
-    fpeer->ip      = 0;
-    fpeer->connect = 0;
+    fpeer->port      = 0;
+    fpeer->ip        = 0;
+    fpeer->connect   = 0;
+    fpeer->version   = serv->assoc->version;
     s_repl (&fpeer->server, NULL);
-    fpeer->uin     = event->uin;
-    fpeer->len     = bytes;
-    fpeer->done    = 0;
-    fpeer->close   = &PeerFileDispatchDClose;
-    IMIntMsg (cont, list, NOW, STATUS_OFFLINE, INT_FILE_ACKING, "", extra);
+    fpeer->cont      = cont;
+    fpeer->len       = opt_bytes;
+    fpeer->done      = 0;
+    fpeer->close     = &PeerFileDispatchDClose;
+    fpeer->reconnect = &TCPDispatchReconn;
+
+    opt = OptSetVals (NULL, CO_BYTES, opt_bytes, CO_MSGTEXT, opt_files, 0);
+    IMIntMsg (cont, serv, NOW, STATUS_OFFLINE, INT_FILE_ACKING, "", opt);
     
     return TRUE;
 }
@@ -155,7 +200,7 @@ BOOL PeerFileAccept (Connection *peer, UWORD status, UDWORD port)
     Connection *flist, *fpeer;
     
     flist = PeerFileCreate (peer->parent->parent);
-    fpeer = ConnectionFind (TYPE_FILEDIRECT, peer->uin, flist);
+    fpeer = ConnectionFind (TYPE_FILEDIRECT, peer->cont, flist);
     
     if (!flist || !fpeer || !port || (status == TCP_ACK_REFUSE))
     {
@@ -170,26 +215,22 @@ BOOL PeerFileAccept (Connection *peer, UWORD status, UDWORD port)
     ASSERT_FILEDIRECT(fpeer);
     
     fpeer->connect  = 0;
-    fpeer->type     = TYPE_FILEDIRECT;
-    fpeer->flags    = 0;
     fpeer->our_seq  = 0;
     fpeer->port     = port;
     fpeer->ip       = peer->ip;
     s_repl (&fpeer->server, s_ip (fpeer->ip));
     fpeer->close    = &PeerFileDispatchDClose;
+    fpeer->reconnect = &TCPDispatchReconn;
     
     if (prG->verbose)
-        M_printf (i18n (2068, "Opening file transfer connection to %s:%ld... \n"),
-                  fpeer->server, fpeer->port);
+        rl_printf (i18n (2520, "Opening file transfer connection to %s:%s%ld%s... \n"),
+                  s_wordquote (fpeer->server), COLQUOTE, fpeer->port, COLNONE);
 
     TCPDispatchConn (fpeer);
     
     return 1;
 }
 
-#define FAIL(x) { err = x; break; }
-#define PeerFileClose TCPClose
-
 /*
  * Close a file listener.
  */
@@ -206,7 +247,8 @@ static void PeerFileDispatchDClose (Connection *fpeer)
 {
     fpeer->connect = 0;
     PeerFileClose (fpeer);
-    R_resetprompt ();
+    ConnectionD (fpeer);
+    ReadLinePromptReset ();
 }
 
 /*
@@ -218,7 +260,7 @@ static void PeerFileIODispatchClose (Connection *ffile)
         close (ffile->sok);
     ffile->sok = -1;
     ffile->connect = 0;
-    R_resetprompt ();
+    ReadLinePromptReset ();
 }
 
 /*
@@ -228,9 +270,10 @@ void PeerFileDispatch (Connection *fpeer)
 {
     Contact *cont;
     Packet *pak;
-    UDWORD err = 0;
+    int err = 0;
     
     ASSERT_FILEDIRECT (fpeer);
+    assert (fpeer->cont);
     
     if (!(pak = UtilIOReceiveTCP (fpeer)))
         return;
@@ -238,12 +281,11 @@ void PeerFileDispatch (Connection *fpeer)
     if (prG->verbose & DEB_PACKTCP)
         TCPPrint (pak, fpeer, FALSE);
 
-    cont = ContactUIN (fpeer->parent->parent, fpeer->uin);
-    assert (cont);
+    cont = fpeer->cont;
 
     switch (PacketRead1 (pak))
     {
-        char *name, *text;
+        strc_t name, text;
         UDWORD len, off, nr, speed;
 
         case 0:
@@ -251,47 +293,43 @@ void PeerFileDispatch (Connection *fpeer)
             nr   = PacketRead4 (pak); /* COUNT */
             len  = PacketRead4 (pak); /* BYTES */
             speed= PacketRead4 (pak); /* SPEED */
-            name = PacketReadLNTS (pak); /* NICK  */
+            name = PacketReadL2Str (pak, NULL); /* NICK  */
             PacketD (pak);
             
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2161, "Receiving %ld files with together %ld bytes at speed %lx from %s.\n"),
-                     nr, len, speed, c_in_to (name, cont));
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2161, "Receiving %ld files with total size %ld bytes at speed %lx from %s.\n"),
+                     nr, len, speed, ConvFromCont (name, cont));
             
             if (len != fpeer->len)
             {
-                M_printf ("FIXME: byte len different than in file request: requested %ld, sending %ld.\n",
+                rl_printf ("FIXME: byte len different than in file request: requested %ld, sending %ld.\n",
                          fpeer->len, len);
                 fpeer->len = len;
             }
             
             pak = PeerPacketC (fpeer, 1);
             PacketWrite4 (pak, 64);
-            PacketWriteLNTS (pak, "my Nick0");
+            PacketWriteLNTS (pak, cont->nick);
             PeerPacketSend (fpeer, pak);
             PacketD (pak);
-            
-            free (name);
             return;
         
         case 1:
-            speed= PacketRead4 (pak); /* SPEED */
-            name = PacketReadLNTS (pak); /* NICK  */
+            speed = PacketRead4 (pak); /* SPEED */
+            name  = PacketReadL2Str (pak, NULL); /* NICK  */
             PacketD (pak);
             
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2170, "Sending with speed %lx to %s.\n"), speed, c_in_to (name, cont));
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2170, "Sending file at speed %lx to %s.\n"), speed, s_wordquote (ConvFromCont (name, cont)));
             
             fpeer->our_seq = 1;
-            QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->uin);
-            
-            free (name);
+            QueueRetry (fpeer, QUEUE_PEER_FILE, cont);
             return;
             
         case 2:
                    PacketRead1 (pak); /* EMPTY */
-            name = PacketReadLNTS (pak);
-            text = PacketReadLNTS (pak);
+            name = PacketReadL2Str (pak, NULL);
+            text = PacketReadL2Str (pak, NULL);
             len  = PacketRead4 (pak);
                    PacketRead4 (pak); /* EMPTY */
                    PacketRead4 (pak); /* SPEED */
@@ -306,14 +344,14 @@ void PeerFileDispatch (Connection *fpeer)
 
                 assert (ffile);
                 pos = snprintf (buf, sizeof (buf), "%sfiles" _OS_PATHSEPSTR "%ld" _OS_PATHSEPSTR,
-                                PrefUserDir (prG), fpeer->uin);
-                snprintf (buf + pos, sizeof (buf) - pos, "%s", c_in_to (name, cont));
+                                PrefUserDir (prG), cont->uin);
+                snprintf (buf + pos, sizeof (buf) - pos, "%s", ConvFromCont (name, cont));
                 for (p = buf + pos; *p; p++)
                     if (*p == '/')
                         *p = '_';
                 finfo.st_size = 0;
                 if (!stat (buf, &finfo))
-                    if (finfo.st_size < len)
+                    if ((UDWORD)finfo.st_size < len)
                         off = finfo.st_size;
                 fpeer->assoc = ffile;
                 ffile->sok = open (buf, O_CREAT | O_WRONLY | (off ? O_APPEND : O_TRUNC), 0660);
@@ -323,17 +361,15 @@ void PeerFileDispatch (Connection *fpeer)
                     if (rc == ENOENT)
                     {
                         mkdir (s_sprintf ("%sfiles", PrefUserDir (prG)), 0700);
-                        mkdir (s_sprintf ("%sfiles" _OS_PATHSEPSTR "%ld", PrefUserDir (prG), fpeer->uin), 0700);
+                        mkdir (s_sprintf ("%sfiles" _OS_PATHSEPSTR "%ld", PrefUserDir (prG), cont->uin), 0700);
                         ffile->sok = open (buf, O_CREAT | O_WRONLY | (off ? O_APPEND : O_TRUNC), 0660);
                     }
                     if (ffile->sok == -1)
                     {
-                        M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                        M_printf (i18n (2083, "Cannot open file %s: %s (%d).\n"),
+                        rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                        rl_printf (i18n (2083, "Cannot open file %s: %s (%d).\n"),
                                  buf, strerror (rc), rc);
-                        ConnectionClose (fpeer);
-                        free (name);
-                        free (text);
+                        ConnectionD (fpeer);
                         return;
                     }
                 }
@@ -342,9 +378,9 @@ void PeerFileDispatch (Connection *fpeer)
                 ffile->done = off;
                 ffile->close = &PeerFileIODispatchClose;
 
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_printf (i18n (2162, "Receiving file %s (%s) with %ld bytes as %s.\n"),
-                         name, text, len, buf);
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_printf (i18n (2162, "Receiving file %s (%s) with %ld bytes as %s.\n"),
+                         name->txt, text->txt, len, buf);
             }
             pak = PeerPacketC (fpeer, 3);
             PacketWrite4 (pak, off);
@@ -353,8 +389,6 @@ void PeerFileDispatch (Connection *fpeer)
             PacketWrite4 (pak, 1);
             PeerPacketSend (fpeer, pak);
             PacketD (pak);
-            free (name);
-            free (text);
             return;
 
         case 3:
@@ -364,16 +398,16 @@ void PeerFileDispatch (Connection *fpeer)
             nr  = PacketRead4 (pak); /* NR */
             PacketD (pak);
             
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2163, "Sending file %ld at offset %ld.\n"),
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2163, "Sending file %ld at offset %ld.\n"),
                      nr, off);
             
             err = lseek (fpeer->assoc->sok, off, SEEK_SET);
             if (err == -1)
             {
                 err = errno;
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_printf (i18n (2084, "Error while seeking to offset %ld: %s (%ld).\n"),
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_printf (i18n (2084, "Error while seeking to offset %ld: %s (%d).\n"),
                          off, strerror (err), err);
                 TCPClose (fpeer);
                 return;
@@ -381,75 +415,85 @@ void PeerFileDispatch (Connection *fpeer)
             fpeer->assoc->done = off;
             fpeer->assoc->connect = CONNECT_OK;
             
-            QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->uin);
+            QueueRetry (fpeer, QUEUE_PEER_FILE, cont);
             return;
             
         case 4:
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2169, "File transfer aborted by peer (%d).\n"),
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2169, "File transfer aborted by peer (%d).\n"),
                      PacketRead1 (pak));
             PacketD (pak);
             PeerFileClose (fpeer);
             return;
 
         case 5:
-            M_printf ("FIXME: Ignoring speed change to %d.\n",
+            rl_printf ("FIXME: Ignoring speed change to %d.\n",
                      PacketRead1 (pak));
             PacketD (pak);
             return;
 
         case 6:
-            len = write (fpeer->assoc->sok, pak->data + 1, pak->len - 1);
-            if (len != pak->len - 1)
+            if (fpeer->assoc->done + pak->len - 1 > fpeer->assoc->len)
             {
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_print  (i18n (2164, "Error writing to file.\n"));
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_printf (i18n (2165, "The peer sent more bytes (%ld) than the file length (%ld).\n"),
+                         fpeer->assoc->done + pak->len - 1, fpeer->assoc->len);
+                PacketD (pak);
+                TCPClose (fpeer);
+                return;
+            }
+            if (pak->len <= 1)
+            {
+                PacketD (pak);
+                return;
+            }
+            len = write (fpeer->assoc->sok, pak->data + 1, pak->len - 1);
+            if (len + 1 != pak->len)
+            {
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_printf (i18n (9999, "Error writing to file (%lu bytes written out of %u).\n"), len, pak->len - 1);
                 PacketD (pak);
                 TCPClose (fpeer);
                 return;
             }
             fpeer->assoc->done += len;
-            if (fpeer->assoc->done > fpeer->assoc->len)
+            if (fpeer->assoc->len == fpeer->assoc->done)
             {
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_printf (i18n (2165, "The peer sent more bytes (%ld) than file length (%ld).\n"),
-                         fpeer->assoc->done, fpeer->assoc->len);
-                PacketD (pak);
-                TCPClose (fpeer);
-                return;
-            }
-            else if (fpeer->assoc->len == fpeer->assoc->done)
-            {
-                R_resetprompt ();
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_print  (i18n (2166, "Finished receiving file.\n"));
+                ReadLinePromptReset ();
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_print  (i18n (2166, "Finished receiving file.\n"));
+                fsync (fpeer->assoc->sok);
+                close (fpeer->assoc->sok);
+                fpeer->assoc->sok = -1;
             }
             else if (fpeer->assoc->len)
             {
-                R_settimepromptf ("[" COLINCOMING "%ld %02d%%" COLNONE "] " COLSERVER "%s" COLNONE "",
-                              fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
-                              i18n (1040, "mICQ> "));
+                ReadLinePromptUpdate (s_sprintf ("[%s%ld %02d%%%s] %s%s",
+                              COLINCOMING, fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
+                              COLNONE, COLSERVER, i18n (2467, "mICQ>")));
             }
             PacketD (pak);
             return;
         default:
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_print  (i18n (2167, "Error - unknown packet.\n"));
-            M_print  (s_dump (pak->data, pak->len));
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_print  (i18n (2167, "Error - unknown packet.\n"));
+            rl_print  (s_dump (pak->data, pak->len));
             PacketD (pak);
             PeerFileClose (fpeer);
     }
     if ((prG->verbose & DEB_TCP) && err)
     {
-        M_printf ("%s %s: %ld\n", s_now, i18n (2029, "Protocol error on peer-to-peer connection"), err);
+        rl_printf ("%s %s: %d\n", s_now, i18n (2029, "Protocol error on peer-to-peer connection"), err);
         PeerFileClose (fpeer);
     }
 }
 
-void PeerFileDispatchW (Connection *fpeer)
+static void PeerFileDispatchW (Connection *fpeer)
 {
     Packet *pak = fpeer->outgoing;
     
+    ASSERT_FILEDIRECT(fpeer);
+    
     fpeer->outgoing = NULL;
     fpeer->connect = CONNECT_OK | CONNECT_SELECT_R;
     fpeer->dispatch = &PeerFileDispatch;
@@ -458,11 +502,15 @@ void PeerFileDispatchW (Connection *fpeer)
     if (!UtilIOSendTCP (fpeer, pak))
         TCPClose (fpeer);
     
-    QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->uin);
+    QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->cont);
 }
 
-BOOL PeerFileError (Connection *fpeer, UDWORD rc, UDWORD flags)
+static BOOL PeerFileError (Connection *fpeer, UDWORD rc, UDWORD flags)
 {
+    Contact *cont = fpeer->cont;
+    
+    ASSERT_FILEDIRECT(fpeer);
+    
     switch (rc)
     {
         case EPIPE:
@@ -476,9 +524,9 @@ BOOL PeerFileError (Connection *fpeer, UDWORD rc, UDWORD flags)
                 fpeer->dispatch = &PeerFileDispatchW;
                 fpeer->assoc->connect = CONNECT_OK | 1;
                 if (fpeer->assoc->len)
-                    R_settimepromptf ("[" COLCONTACT "%ld:%02d%%" COLNONE "] " COLSERVER "%s" COLNONE "",
-                                  fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
-                                  i18n (1040, "mICQ> "));
+                    ReadLinePromptUpdate (s_sprintf ("[%s%ld:%02d%%%s] %s%s",
+                                  COLCONTACT, fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
+                                  COLNONE, COLSERVER, i18n (2467, "mICQ>")));
                 return 1;
             }
     }
@@ -492,7 +540,7 @@ void PeerFileResend (Event *event)
     Packet *pak;
     Event *event2;
     int rc;
-    const char *e_msg_text = "<>";
+    const char *opt_text;
     
     if (!fpeer)
     {
@@ -502,23 +550,24 @@ void PeerFileResend (Event *event)
     
     ASSERT_FILEDIRECT (fpeer);
 
-    cont = ContactUIN (event->conn, event->uin);
+    cont = event->cont;
     assert (cont);
     
-    e_msg_text = ExtraGetS (event->extra, EXTRA_MESSAGE);
+    if (!OptGetStr (event->opt, CO_FILENAME, &opt_text))
+        opt_text = "";
 
-    if (event->attempts >= MAX_RETRY_P2P_ATTEMPTS + 4 || (!event->pak && !event->seq))
+    if (event->attempts >= MAX_RETRY_P2PFILE_ATTEMPTS || (!event->pak && !event->seq))
     {
-        M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-        M_printf (i18n (2168, "File transfer #%ld (%s) dropped after %ld attempts because of timeout.\n"),
-                 event->seq, e_msg_text, event->attempts);
+        rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+        rl_printf (i18n (2168, "File transfer #%ld (%s) dropped after %ld attempts because of timeout.\n"),
+                 event->seq, opt_text, event->attempts);
         TCPClose (fpeer);
     }
     else if (!(fpeer->connect & CONNECT_MASK))
     {
-        M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-        M_printf (i18n (2072, "File transfer #%ld (%s) canceled because of closed connection.\n"),
-                 event->seq, e_msg_text);
+        rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+        rl_printf (i18n (2072, "File transfer #%ld (%s) canceled because of closed connection.\n"),
+                 event->seq, opt_text);
     }
     else if (~fpeer->connect & CONNECT_OK)
     {
@@ -554,24 +603,24 @@ void PeerFileResend (Event *event)
         ffile = ConnectionClone (fpeer, TYPE_FILE);
         fpeer->assoc = ffile;
 
-        if (stat (e_msg_text, &finfo))
+        if (stat (opt_text, &finfo))
         {
             rc = errno;
-            M_printf (i18n (2071, "Couldn't stat file %s: %s (%d)\n"),
-                      e_msg_text, strerror (rc), rc);
+            rl_printf (i18n (2071, "Couldn't stat file %s: %s (%d)\n"),
+                      s_wordquote (opt_text), strerror (rc), rc);
         }
         ffile->len = finfo.st_size;
 
-        ffile->sok = open (e_msg_text, O_RDONLY);
+        ffile->sok = open (opt_text, O_RDONLY);
         if (ffile->sok == -1)
         {
             int rc = errno;
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2083, "Cannot open file %s: %s (%d).\n"),
-                      e_msg_text, strerror (rc), rc);
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2083, "Cannot open file %s: %s (%d).\n"),
+                      opt_text, strerror (rc), rc);
             TCPClose (fpeer);
-            ConnectionClose (ffile);
-            ConnectionClose (fpeer);
+            ConnectionD (ffile);
+            ConnectionD (fpeer);
             return;
         }
         ffile->close = &PeerFileIODispatchClose;
@@ -593,9 +642,9 @@ void PeerFileResend (Event *event)
         if (len == -1)
         {
             len = errno;
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2086, "Error while reading file %s: %s (%d).\n"),
-                      e_msg_text, strerror (len), len);
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2086, "Error while reading file %s: %s (%d).\n"),
+                      opt_text, strerror (len), len);
             TCPClose (fpeer);
         }
         else
@@ -606,37 +655,38 @@ void PeerFileResend (Event *event)
             PeerPacketSend (fpeer, pak);
             PacketD (pak);
 
-            if (len == 2048)
+            if (len > 0)
             {
                 if (fpeer->assoc->len)
-                    R_settimepromptf ("[" COLCONTACT "%ld %02d%%" COLNONE "] " COLSERVER "%s" COLNONE "",
-                                  fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
-                                  i18n (1040, "mICQ> "));
+                    ReadLinePromptUpdate (s_sprintf ("[%s%ld %02d%%%s] %s%s",
+                                  COLCONTACT, fpeer->assoc->done, (int)((100.0 * fpeer->assoc->done) / fpeer->assoc->len),
+                                  COLNONE, COLSERVER, i18n (2467, "mICQ>")));
                 else
-                    R_settimepromptf ("[" COLCONTACT "%ld" COLNONE "] " COLSERVER "%s" COLNONE "",
-                                  fpeer->assoc->done, i18n (1040, "mICQ> "));
+                    ReadLinePromptUpdate (s_sprintf ("[%s%ld%s] %s%s",
+                                  COLCONTACT, fpeer->assoc->done,
+                                  COLNONE, COLSERVER, i18n (2467, "mICQ>")));
                 event->attempts = 0;
                 QueueEnqueue (event);
                 return;
             }
 
-            R_resetprompt ();
-            M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-            M_printf (i18n (2087, "Finished sending file %s.\n"), e_msg_text);
-            ConnectionClose (fpeer->assoc);
+            ReadLinePromptReset ();
+            rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+            rl_printf (i18n (2087, "Finished sending file %s.\n"), opt_text);
+            ConnectionD (fpeer->assoc);
             fpeer->our_seq++;
             event2 = QueueDequeue (fpeer, QUEUE_PEER_FILE, fpeer->our_seq);
             if (event2)
             {
                 QueueEnqueue (event2);
-                QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->uin);
+                QueueRetry (fpeer, QUEUE_PEER_FILE, fpeer->cont);
                 return;
             }
             else
             {
-                M_printf ("%s " COLCONTACT "%*s" COLNONE " ", s_now, uiG.nick_len + s_delta (cont->nick), cont->nick);
-                M_printf (i18n (2088, "Finished sending all %d files.\n"), fpeer->our_seq - 1);
-                ConnectionClose (fpeer);
+                rl_printf ("%s %s%*s%s ", s_now, COLCONTACT, uiG.nick_len + s_delta (cont->nick), cont->nick, COLNONE);
+                rl_printf (i18n (2088, "Finished sending all %d files.\n"), fpeer->our_seq - 1);
+                ConnectionD (fpeer);
             }
         }
     }
