@@ -12,12 +12,21 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  * License for more details.
  *
+ * In addition, as a special exception permission is granted to link the
+ * code of this release of mICQ with the OpenSSL project's "OpenSSL"
+ * library, and distribute the linked executables.  You must obey the GNU
+ * General Public License in all respects for all of the code used other
+ * than "OpenSSL".  If you modify this file, you may extend this exception
+ * to your version of the file, but you are not obligated to do so.  If you
+ * do not wish to do so, delete this exception statement from your version
+ * of this file.
+ *
  * You should have received a copy of the GNU General Public License
  * along with this package; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  * 02111-1307, USA.
  *
- * $Id: util_str.c,v 1.35.2.18 2003/10/05 01:46:14 kuhlmann Exp $
+ * $Id: util_str.c,v 1.75 2004/12/05 14:55:31 kuhlmann Exp $
  */
 
 #include "micq.h"
@@ -34,14 +43,228 @@
 #if HAVE_WINSOCK2_H
 #include <winsock2.h>
 #endif
-#include <string.h>
+#if HAVE_WCTYPE_H
+#include <wctype.h>
+#endif
 #include <ctype.h>
 #include "util_str.h"
-#include "preferences.h"
-#include "contact.h"
 #include "conv.h"
-#include "session.h" /* conn->contacts */
-#include "util.h"
+#include "contact.h"
+#include "preferences.h"
+
+/*
+ * Initialize a dynamically allocated string.
+ *
+ * Returns NULL if not enough memory.
+ */
+str_t s_init (str_t str, const char *init, size_t add)
+{
+    UDWORD initlen, size;
+    
+    initlen = strlen (init);
+    size = ((initlen + add) | 0x7f) + 1;
+    
+    if (!str->txt || size > str->max)
+    {
+        if (str->txt && str->max)
+            free (str->txt);
+        str->txt = malloc (size);
+        if (!str->txt)
+        {
+            str->max = str->len = 0;
+            str->txt = "\0";
+            return NULL;
+        }
+        str->max = size;
+    }
+    str->len = initlen;
+    memcpy (str->txt, init, initlen + 1);
+    return str;
+}
+
+/*
+ * Increase storage size of dynamically allocated string.
+ *
+ * Returns NULL if not enough memory.
+ */
+str_t s_blow (str_t str, size_t len)
+{
+    int newlen;
+    char *tmp;
+    
+    newlen = 1 + ((str->max + len - 1) | 0x7f);
+    
+    if (!str->txt || !str->max)
+        tmp = malloc (newlen);
+    else
+        tmp = realloc (str->txt, newlen);
+    
+    if (!tmp)
+    {
+        if (!str->txt || !str->max)
+        {
+            str->max = str->len = 0;
+            str->txt = "\0";
+        }
+        return NULL;
+    }
+    str->txt = tmp;
+    str->max = newlen;
+    return str;
+}
+
+
+/*
+ * Appends a byte to a dynamically allocated string.
+ */
+str_t s_catc (str_t str, char add)
+{
+    if (str->len + 2 > str->max)
+        if (!s_blow (str, str->len + 2 - str->max))
+            return NULL;
+
+    str->txt[str->len++] = add;
+    str->txt[str->len] = '\0';
+    return str;
+}
+
+/*
+ * Appends bytes to a dynamically allocated string.
+ */
+str_t s_catn (str_t str, const char *add, size_t len)
+{
+    if (str->len + len + 2 > str->max)
+        if (!s_blow (str, str->len + len + 2 - str->max))
+            return NULL;
+
+    memcpy (str->txt + str->len, add, len);
+    str->len += len;
+    str->txt[str->len] = '\0';
+    return str;
+}
+
+/*
+ * Appends a string to a dynamically allocated string.
+ */
+str_t s_cat (str_t str, const char *add)
+{
+    int addlen = strlen (add);
+    
+    if (str->len + addlen + 2 > str->max)
+        if (!s_blow (str, str->len + addlen + 2 - str->max))
+            return NULL;
+    
+    memcpy (str->txt + str->len, add, addlen + 1);
+    str->len += addlen;
+    return str;
+}
+
+
+/*
+ * Appends a formatted string to a dynamically allocated string.
+ */
+str_t s_catf (str_t str, const char *fmt, ...)
+{
+    va_list args;
+    size_t add = 128;
+    int rc;
+    
+    if (str->max < add || str->len + 2 >= str->max)
+        s_blow (str, add);
+
+    while (1)
+    {
+        if (str->len + 2 >= str->max)
+            return NULL;
+        str->txt[str->max - 2] = '\0';
+        va_start (args, fmt);
+        rc = vsnprintf (str->txt + str->len, str->max - str->len - 1, fmt, args);
+        va_end (args);
+        if (rc >= 0 && rc < str->max - str->len - 2 && !str->txt[str->max - 2])
+        {
+            str->len += strlen (str->txt + str->len);
+            return str;
+        }
+        add = (rc > 0 ? rc - str->max + str->len + 5 : str->max);
+        if (!s_blow (str, add))
+        {
+            str->txt[str->max - 2] = '\0';
+            str->len = str->max - 2;
+            return str;
+        }
+    }
+    return str;
+}
+
+/*
+ * Insert a string into a dynamically allocated string.
+ */
+str_t s_insn (str_t str, size_t pos, const char *ins, size_t len)
+{
+    if (pos > str->len || pos < 0)
+        return NULL;
+    if (str->len + len + 2 > str->max)
+        if (!s_blow (str, str->len + len + 2 - str->max))
+            return NULL;
+    
+    memmove (str->txt + pos + len, str->txt + pos, str->len - pos + 1);
+    memcpy (str->txt + pos, ins, len);
+    str->len += len;
+    return str;
+}
+
+/*
+ * Insert a byte into a dynamically allocated string.
+ */
+str_t s_insc (str_t str, size_t pos, char ins)
+{
+    if (pos > str->len || pos < 0)
+        return NULL;
+    if (str->len + 3 > str->max)
+        if (!s_blow (str, str->len + 3 - str->max))
+            return NULL;
+    
+    memmove (str->txt + pos + 1, str->txt + pos, str->len - pos + 1);
+    str->txt[pos] = ins;
+    str->len++;
+    return str;
+}
+
+/*
+ * Delete a byte from a dynamically allocated string.
+ */
+str_t s_delc (str_t str, size_t pos)
+{
+    if (pos > str->len || pos < 0)
+        return NULL;
+    memmove (str->txt + pos, str->txt + pos + 1, str->len - pos);
+    str->len--;
+    return str;
+}
+
+/*
+ * Delete bytes from a dynamically allocated string.
+ */
+str_t s_deln (str_t str, size_t pos, size_t len)
+{
+    if (pos + len > str->len)
+        return NULL;
+    memmove (str->txt + pos, str->txt + pos + len, str->len - pos);
+    str->len -= len;
+    return str;
+}
+
+
+/*
+ * Frees a dynamically allocated string.
+ */
+void s_done (str_t str)
+{
+    if (str->txt && str->max)
+        free (str->txt);
+    str->txt = "\0";
+    str->len = str->max = 0;
+}
 
 /*
  * Return a static formatted string.
@@ -49,10 +272,10 @@
 const char *s_sprintf (const char *fmt, ...)
 {
     static char *buf = NULL;
-    static UDWORD size = 0;
+    static int size = 0;
     va_list args;
     char *nbuf;
-    UDWORD rc;
+    int rc, nsize;
 
     if (!buf)
         buf = calloc (1, size = 1024);
@@ -64,83 +287,20 @@ const char *s_sprintf (const char *fmt, ...)
         rc = vsnprintf (buf, size, fmt, args);
         va_end (args);
         
-        if (rc != -1 && rc < size && !buf[size - 2])
+        if (rc >= 0 && rc < size && !buf[size - 2])
             break;
 
-        nbuf = malloc (size + 1024);
+        nsize = (rc > 0 ? rc + 5 : size * 2);
+        nbuf = malloc (nsize);
         if (!nbuf)
             break;
         free (buf);
         buf = nbuf;
-        size += 1024;
+        size = nsize;
     }
     return buf;
 }
 
-/*
- * Appends to a dynamically allocated string.
- */
-char *s_cat (char *str, UDWORD *size, const char *add)
-{
-    UDWORD nsize;
-    char *nstr;
-    
-    if (!str)
-        str = strdup ("");
-    if (!size)
-        return str;
-    
-    nsize = strlen (nstr = str) + strlen (add) + 1;
-    if (nsize > *size)
-    {
-        nstr = realloc (str, nsize += 64);
-        if (nstr)
-            *size = nsize;
-        else
-            return str;
-    }
-    strcat (nstr, add);
-    return nstr;
-}
-
-
-/*
- * Appends to a dynamically allocated string.
- */
-char *s_catf (char *str, UDWORD *size, const char *fmt, ...)
-{
-    va_list args;
-    UDWORD nsize;
-    char *nstr;
-    int rc, strln;
-    
-    if (!size)
-        return str;
-    if (!str || !*size)
-        str = calloc (*size = 1024, 1);
-    strln = strlen (str);
-    
-    while (1)
-    {
-        str[*size - 2] = '\0';
-        va_start (args, fmt);
-        rc = vsnprintf (str + strln, *size - strln, fmt, args);
-        va_end (args);
-        if (rc >= 0 && rc < *size - strln && !str[*size - 2])
-            return str;
-        nsize = (rc > 0 ? rc + strln + 5 : 2 * *size);
-        nstr = realloc (str, nsize);
-        if (!nstr)
-        {
-            str[*size - 2] = '\0';
-            return str;
-        }
-        str = nstr;
-        *size = nsize;
-    }
-    return str;
-}
-
 /*
  * Return a static string consisting of the given IP.
  */
@@ -260,22 +420,23 @@ const char *s_msgtok (char *txt)
  */
 const char *s_ind (const char *str)
 {
-    static char *t = NULL;
-    static UDWORD size = 0;
+    static str_s t;
     UDWORD cnt = 5;
     const char *p;
     char *q;
     
+    if (!str || !*str)
+        return "";
+
     for (p = str; *p; p++, cnt++)
         if (*p == '\n')
             cnt += 2;
-    if (!t || !str || !*str)
-        t = calloc (1, size = 100);
-    if (size < cnt)
-        t = realloc (t, cnt);
-    if (!t || !str || !*str)
-        return str;
-    q = t;
+    
+    s_init (&t, "", cnt);
+    if (t.max < cnt)
+        return "<out of memory>";
+
+    q = t.txt;
     *q++ = ' ';
     *q++ = ' ';
     for (p = str; *p; )
@@ -286,7 +447,7 @@ const char *s_ind (const char *str)
                 *q++ = ' ';
             }
     *q = '\0';
-    return t;
+    return t.txt;
 }
 
 /*
@@ -335,81 +496,117 @@ UDWORD s_offset  (const char *str, UDWORD offset)
     return off;
 }
 
-#ifdef ENABLE_UTF8
+/*
+ * Split off a certain amount of recoded bytes
+ */
+strc_t s_split (const char **input, UBYTE enc, int len)
+{
+    static str_s str = { NULL, 0, 0 };
+    str_s in = { NULL, 0, 0 };
+    strc_t tr;
+    int off, offt, offold, offnl, offin, offmax;
+    UDWORD ucs;
+    
+    in.txt = (char *) *input;
+    in.len = strlen (*input);
+    off = offt = offnl = offin = offmax = 0;
+    
+    while (in.txt[off])
+    {
+        offold = off;
+        ucs = ConvGetUTF8 (&in, &off);
+        tr = ConvToLen (in.txt + offold, enc, off - offold);
+        if (ucs == '\r')
+            offnl = offold;
+        if (!(ucs & ~0xff) && isspace (ucs & 0xff))
+            offin = offold;
+        if (tr->len > len)
+        {
+            off = offold;
+            break;
+        }
+        if (ucs == '\n')
+            offnl = off;
+        if (!iswalnum (ucs))
+            offin = off;
+        len -= tr->len;
+        if (!in.txt[off])
+            offmax = off;
+    }
+    if (offmax)
+        off = offmax;
+    else if (offnl)
+        off = offnl;
+    else if (offin)
+        off = offin;
+    s_init (&str, "", off + 2);
+    memcpy (str.txt, in.txt, off);
+    str.len = off;
+    str.txt[off] = 0;
+    *input += off;
+    return ConvTo (str.txt, enc);
+}
+
+
 #define noctl(x) ((((x & 0x60) && x != 0x7f)) ? ConvUTF8 (x) : ".")
-#else
-#define noctl(x) ((((x & 0x60) && x != 0x7f)) ? x : '.')
-#endif
 
 /*
  * Hex dump to a string with ASCII.
  */
 const char *s_dump (const UBYTE *data, UWORD len)
 {
-    static char *t = NULL;
-    static UDWORD size = 0;
+    static str_s t;
     UDWORD i, off;
     const unsigned char *d = (const unsigned char *)data;
     
-    if (!t)
-        t = calloc (1, size = 100);
-    *t = '\0';
+    s_init (&t, "", 100);
+
     while (len >= 16)
     {
-        t = s_catf (t, &size, "%02x %02x %02x %02x %02x %02x %02x %02x  "
+        s_catf (&t, "%02x %02x %02x %02x %02x %02x %02x %02x  "
                               "%02x %02x %02x %02x %02x %02x %02x %02x  ",
                     d[0], d[1],  d[2],  d[3],  d[4],  d[5],  d[6],  d[7],
                     d[8], d[9], d[10], d[11], d[12], d[13], d[14], d[15]);
-#ifdef ENABLE_UTF8
-        t = s_cat (t, &size, "\"");
-        t = s_cat (t, &size, noctl  (d[0]));  t = s_cat (t, &size, noctl  (d[1]));
-        t = s_cat (t, &size, noctl  (d[2]));  t = s_cat (t, &size, noctl  (d[3]));
-        t = s_cat (t, &size, noctl  (d[4]));  t = s_cat (t, &size, noctl  (d[5]));
-        t = s_cat (t, &size, noctl  (d[6]));  t = s_cat (t, &size, noctl  (d[7]));
-        t = s_cat (t, &size, " ");
-        t = s_cat (t, &size, noctl  (d[8]));  t = s_cat (t, &size, noctl  (d[9]));
-        t = s_cat (t, &size, noctl (d[10]));  t = s_cat (t, &size, noctl (d[11]));
-        t = s_cat (t, &size, noctl (d[12]));  t = s_cat (t, &size, noctl (d[13]));
-        t = s_cat (t, &size, noctl (d[14]));  t = s_cat (t, &size, noctl (d[15]));
-        t = s_cat (t, &size, "'\n");
-#else
-        t = s_catf (t, &size, "'%c%c%c%c%c%c%c%c %c%c%c%c%c%c%c%c'\n",
-                    noctl  (d[0]), noctl  (d[1]), noctl  (d[2]), noctl  (d[3]),
-                    noctl  (d[4]), noctl  (d[5]), noctl  (d[6]), noctl  (d[7]),
-                    noctl  (d[8]), noctl  (d[9]), noctl (d[10]), noctl (d[11]),
-                    noctl (d[12]), noctl (d[13]), noctl (d[14]), noctl (d[15]));
-#endif
+        s_catc (&t, '\"');
+        s_cat (&t, noctl  (d[0]));  s_cat (&t, noctl  (d[1]));
+        s_cat (&t, noctl  (d[2]));  s_cat (&t, noctl  (d[3]));
+        s_cat (&t, noctl  (d[4]));  s_cat (&t, noctl  (d[5]));
+        s_cat (&t, noctl  (d[6]));  s_cat (&t, noctl  (d[7]));
+        s_cat (&t, " ");
+        s_cat (&t, noctl  (d[8]));  s_cat (&t, noctl  (d[9]));
+        s_cat (&t, noctl (d[10]));  s_cat (&t, noctl (d[11]));
+        s_cat (&t, noctl (d[12]));  s_cat (&t, noctl (d[13]));
+        s_cat (&t, noctl (d[14]));  s_cat (&t, noctl (d[15]));
+        s_catc (&t, '"');
+        s_catc (&t, '\n');
         len -= 16;
         d += 16;
     }
     if (!len)
-        return (t);
+        return (t.txt);
     for (off = i = 0; i <= 16; i++)
     {
         if (len)
         {
-            t = s_catf (t, &size, "%02x ", *d++);
+            s_catf (&t, "%02x ", *d++);
             len--;
             off++;
         }
         else if (i != 16)
-            t = s_catf (t, &size, "   ");
+            s_catf (&t, "   ");
         if (i == 7)
-            t = s_catf (t, &size, " ");
+            s_catc (&t, ' ');
         
     }
-    t = s_catf (t, &size, " '");
+    s_catc (&t, ' ');
+    s_catc (&t, '\'');
     while (off)
     {
-#ifdef ENABLE_UTF8
-        t = s_catf (t, &size, "%s", noctl (*(d - off)));
-#else
-        t = s_catf (t, &size, "%c", noctl (*(d - off)));
-#endif
+        s_cat (&t, noctl (*(d - off)));
         off--;
     }
-    t = s_catf (t, &size, "'\n");
-    return t;
+    s_catf (&t, "'\n");
+    return t.txt;
 }
 
 /*
@@ -417,17 +614,16 @@ const char *s_dump (const UBYTE *data, UWORD len)
  */
 const char *s_dumpnd (const UBYTE *data, UWORD len)
 {
-    static char *t = NULL;
-    static UDWORD size = 0;
+    static str_s t;
     UDWORD i;
     const unsigned char *d = (const unsigned char *)data;
     
-    if (t)
-        *t = 0;
+    s_init (&t, "", 100);
+
     while (len > 16)
     {
-        t = s_catf (t, &size, "%02x %02x %02x %02x %02x %02x %02x %02x  "
-                              "%02x %02x %02x %02x %02x %02x %02x %02x\\\n",
+        s_catf (&t, "%02x %02x %02x %02x %02x %02x %02x %02x  "
+                    "%02x %02x %02x %02x %02x %02x %02x %02x\\\n",
                     d[0], d[1],  d[2],  d[3],  d[4],  d[5],  d[6],  d[7],
                     d[8], d[9], d[10], d[11], d[12], d[13], d[14], d[15]);
         len -= 16;
@@ -437,24 +633,24 @@ const char *s_dumpnd (const UBYTE *data, UWORD len)
     {
         for (i = 0; i < len && i < 8; i++)
         {
-            t = s_catf (t, &size, "%02x ", *d++);
+            s_catf (&t, "%02x ", *d++);
         }
-        t = s_catf (t, &size, "\n");
+        s_catc (&t, '\n');
         len -= i;
     }
     for (i = 0; i < 10; i++)
     {
         if (len)
         {
-            t = s_catf (t, &size, "%02x ", *d++);
+            s_catf (&t, "%02x ", *d++);
             len--;
         }
         else
-            t = s_catf (t, &size, "   ");
+            s_cat (&t, "   ");
         if (i == 7)
-            t = s_catf (t, &size, " ");
+            s_catc (&t, ' ');
     }
-    return t;
+    return t.txt;
 }
 
 /*
@@ -483,278 +679,6 @@ const char *s_realpath (const char *path)
     return path;
 }
 
-/*
- * s_parse* - find a parameter of given type in string.
- * input is advanced to point after the parsed argument,
- *   or to the next non-whitespace or end of string if not found.
- * parsed is cleared (0, NULL) if no suitable argument found.
- */
-
-/*
- * Try to find a parameter in the string.
- * Result must NOT be free()d.
- */
-BOOL s_parse_s (char **input, char **parsed, char *sep)
-{
-    static char *t = NULL;
-    char *p = *input, *q;
-    int s = 0;
-    
-    while (*p && strchr (sep, *p))
-        p++;
-    if (*p == '#')
-    {
-        while (*p)
-            p++;
-    }
-    *input = p;
-    if (!*p)
-    {
-        *parsed = NULL;
-        return FALSE;
-    }
-
-    s_repl (&t, p);
-
-    *parsed = q = t;
-    if (!t)
-    {
-        return FALSE;
-    }
-    
-    if (*p == '"')
-    {
-        s = 1;
-        p++;
-    }
-    while (*p)
-    {
-        if (*p == '\\')
-        {
-            p++;
-            if (*p == 'x' && p[1] && p[2])
-            {
-                p++;
-                *q = (*p >= '0' && *p <= '9' ? *p - '0' : *p >= 'a' && *p <= 'f' ? *p - 'a' + 10 : *p - 'A' + 10) << 4;
-                p++;
-                *q = *p >= '0' && *p <= '9' ? *p - '0' : *p >= 'a' && *p <= 'f' ? *p - 'a' + 10 : *p - 'A' + 10;
-                p++, q++;
-                continue;
-            }
-            else if (*p)
-            {
-                *(q++) = *(p++);
-                continue;
-            }
-        }
-        if (*p == '"' && s)
-        {
-            *q = '\0';
-            *input = p + 1;
-            return TRUE;
-        }
-        if (!s && strchr (sep, *p))
-            break;
-        *(q++) = *(p++);
-    }
-    *q = '\0';
-    *input = p;
-    return TRUE;
-}
-
-/*
- * Try to find a nick name or uin in the string.
- * parsed is the Contact entry for this alias,
- * parsedr the "real" entry for this UIN.
- * parsed ->nick will be the nick given,
- * unless the user entered an UIN.
- */
-BOOL s_parsenick_s (char **input, Contact **parsed, char *sep, Contact **parsedr, Connection *serv)
-{
-    ContactGroup *cg;
-    Contact *r;
-    char *p = *input, *t;
-    UDWORD max, l, ll, i;
-    
-    while (*p && strchr (sep, *p))
-        p++;
-    *input = p;
-    if (!*p)
-    {
-        *parsed = NULL;
-        if (parsedr)
-            *parsedr = NULL;
-        return FALSE;
-    }
-    
-    t = NULL;
-    if (s_parse_s (&p, &t, sep))
-    {
-        *parsed = ContactFind (serv->contacts, 0, 0, t, 0);
-        if (*parsed)
-        {
-            if (parsedr)
-            {
-                *parsedr = ContactFind (serv->contacts, 0, (*parsed)->uin, NULL, 0);
-                if (!*parsedr)
-                {
-                    *parsed = NULL;
-                    return FALSE;
-                }
-            }
-            *input = p;
-            return TRUE;
-        }
-    }
-    p = *input;
-
-    if (strchr ("0123456789", *p))
-    {
-        l = 0;
-        if (s_parseint_s (&p, &max, sep))
-        {
-            if ((r = ContactUIN (serv, max)))
-            {
-                *parsed = r;
-                if (parsedr)
-                    *parsedr = r;
-                *input = p;
-                return TRUE;
-            }
-        }
-    }
-    max = 0;
-    *parsed = NULL;
-    ll = strlen (p);
-    cg = serv->contacts;
-    for (i = 0; (r = ContactIndex (cg, i)); i++)
-    {
-        l = strlen (r->nick);
-        if (l > max && l <= ll && (!p[l] || strchr (sep, p[l])) && !strncasecmp (p, r->nick, l))
-        {
-            *parsed = r;
-            max = strlen (r->nick);
-        }
-    }
-    if (max)
-    {
-        if (parsedr)
-        {
-            *parsedr = ContactFind (serv->contacts, 0, (*parsed)->uin, NULL, 0);
-            if (!*parsedr)
-            {
-                *parsed = NULL;
-                return FALSE;
-            }
-        }
-        *input = p + max;
-        return TRUE;
-    }
-    return FALSE;
-}
-
-/*
- * Finds the remaining non-whitespace line, but parses '\'.
- */
-BOOL s_parserem_s (char **input, char **parsed, char *sep)
-{
-    static char *t = NULL;
-    char *p = *input, *q;
-    
-    while (*p && strchr (sep, *p))
-        p++;
-    *input = p;
-    if (!*p)
-    {
-        *parsed = NULL;
-        return FALSE;
-    }
-
-    s_repl (&t, p);
-    *parsed = t;
-    for (q = t ; *p; )
-    {
-        if (*p == '\\')
-        {
-            p++;
-            if (*p == 'x' && p[1] && p[2])
-            {
-                p++;
-                *q = (*p >= '0' && *p <= '9' ? *p - '0' : *p >= 'a' && *p <= 'f' ? *p - 'a' + 10 : *p - 'A' + 10) << 4;
-                p++;
-                *q |= *p >= '0' && *p <= '9' ? *p - '0' : *p >= 'a' && *p <= 'f' ? *p - 'a' + 10 : *p - 'A' + 10;
-                p++, q++;
-            }
-            else if (*p)
-                *(q++) = *(p++);
-            else
-                *(q++) = '\\';
-        }
-        else
-            *(q++) = *(p++);
-    }
-    *q = '\0';
-    *input = p;
-    return TRUE;
-}
-
-/*
- * Try to find a number.
- */
-BOOL s_parseint_s (char **input, UDWORD *parsed, char *sep)
-{
-    char *p = *input;
-    UDWORD nr, sig;
-    
-    while (*p && strchr (sep, *p))
-        p++;
-    if (*p == '#')
-    {
-        while (*p)
-            p++;
-    }
-    *input = p;
-    if (!*p || !strchr ("0123456789+-", *p))
-    {
-        *parsed = 0;
-        return FALSE;
-    }
-
-    nr = 0;
-    sig = 1;
-
-    if (*p == '+')
-        p++;
-    if (*p == '-')
-    {
-        sig = -1;
-        p++;
-    }
-
-    while (*p && *p >= '0' && *p <= '9')
-    {
-        nr = nr * 10 + (*p - '0');
-        p++;
-    }
-    if (!nr && *p == 'x')
-    {
-        p++;
-        while (*p && ((*p >= '0' && *p <= '9') || (*p >= 'a' && *p <= 'f') || (*p >= 'A' && *p <= 'F')))
-        {
-            nr = nr * 16 + (*p >= '0' && *p <= '9' ? *p - '0' : *p >= 'a' && *p <= 'f' ? *p - 'a' + 10 : *p - 'A' + 10);
-            p++;
-        }
-    }
-    if (*p && !strchr (sep, *p))
-    {
-        *parsed = 0;
-        return FALSE;
-    }
-    *input = p;
-    *parsed = nr * sig;
-    return TRUE;
-}
-
 #define SUPERSAFE "%*+-_.:=@/0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
 
 /*
@@ -762,68 +686,130 @@ BOOL s_parseint_s (char **input, UDWORD *parsed, char *sep)
  */
 const char *s_quote (const char *input)
 {
-    static char *t = NULL;
-    static UDWORD size = 0;
+    static str_s t;
     const char *tmp;
     
-    if (!t)
-        t = malloc (size = 32);
-    if (!t || !input || !*input)
+    if (!input || !*input)
         return "\"\"";
     for (tmp = input; *tmp; tmp++)
         if (!strchr (SUPERSAFE, *tmp))
             break;
     if (!*tmp)
         return input;
-    *t = 0;
-    t = s_cat (t, &size, "\"");
+    
+    s_init (&t, "", 100);
+
+    s_catc (&t, '\"');
     for (tmp = input; *tmp; tmp++)
     {
         if (strchr ("\\\"", *tmp))
-            t = s_catf (t, &size, "\\%c", *tmp);
+        {
+            s_catc (&t, '\\');
+            s_catc (&t, *tmp);
+        }
         else if (*tmp & 0xe0)
-            t = s_catf (t, &size, "%c", *tmp);
+            s_catc (&t, *tmp);
         else
-            t = s_catf (t, &size, "\\x%c%c",
-                    (*tmp / 16) <= 9 ? ((UBYTE)*tmp / 16) + '0'
-                                     : ((UBYTE)*tmp / 16) - 10 + 'a',
-                    (*tmp & 15) <= 9 ? (*tmp & 15) + '0'
+        {
+            s_catc (&t, '\\');
+            s_catc (&t, 'x');
+            s_catc (&t, (*tmp / 16) <= 9 ? ((UBYTE)*tmp / 16) + '0'
+                                     : ((UBYTE)*tmp / 16) - 10 + 'a');
+            s_catc (&t, (*tmp & 15) <= 9 ? (*tmp & 15) + '0'
                                      : (*tmp & 15) - 10 + 'a');
+        }
     }
-    t = s_cat (t, &size, "\"");
-    return t;
+    s_catc (&t, '\"');
+    return t.txt;
 }
 
 /*
- * Remove all control characters from string
+ * Quote a string for display. May use quotes to make sure text boundary is clear.
  */
-const char *s_sanitize (const char *input)
+const char *s_cquote (const char *input, const char *color)
 {
-    static char *t = NULL;
-    static UDWORD size = 0;
+    static str_s t;
     const char *tmp;
     
-    if (!t)
-        t = malloc (size = 32);
-    if (!t || !input || !*input)
-        return "";
+    if (!input || !*input)
+        return "\"\"";
+    s_init (&t, "", 100);
+
     for (tmp = input; *tmp; tmp++)
-        if (!(*tmp & 0xe0) || (*tmp == 127))
+        if (!strchr (SUPERSAFE, *tmp))
             break;
     if (!*tmp)
-        return input;
-    *t = 0;
+    {
+        s_cat (&t, color);
+        s_cat (&t, input);
+        s_cat (&t, COLNONE);
+        return t.txt;
+    }
+    
+    s_catc (&t, '\"');
+    s_cat (&t, color);
     for (tmp = input; *tmp; tmp++)
     {
-        if ((*tmp & 0xe0) && (*tmp != 127))
-            t = s_catf (t, &size, "%c", *tmp);
+        if (strchr ("\\\"", *tmp))
+        {
+            s_catc (&t, '\\');
+            s_catc (&t, *tmp);
+        }
+        else if (*tmp & 0xe0)
+            s_catc (&t, *tmp);
         else
-            t = s_catf (t, &size, "\\x%c%c",
-                    (*tmp / 16) <= 9 ? ((UBYTE)*tmp / 16) + '0'
-                                     : ((UBYTE)*tmp / 16) - 10 + 'a',
-                    (*tmp & 15) <= 9 ? (*tmp & 15) + '0'
+        {
+            s_catc (&t, '\\');
+            s_catc (&t, 'x');
+            s_catc (&t, (*tmp / 16) <= 9 ? ((UBYTE)*tmp / 16) + '0'
+                                     : ((UBYTE)*tmp / 16) - 10 + 'a');
+            s_catc (&t, (*tmp & 15) <= 9 ? (*tmp & 15) + '0'
                                      : (*tmp & 15) - 10 + 'a');
+        }
     }
-    return t;
+    s_cat (&t, COLNONE);
+    s_catc (&t, '\"');
+    return t.txt;
+}
+
+/*
+ * Quote a string for message display. Uses quotes only for the empty string.
+ */
+const char *s_mquote (const char *input, const char *color, BOOL allownl)
+{
+    static str_s t;
+    const char *tmp;
+    
+    if (!input || !*input)
+        return "\"\"";
+    s_init (&t, "", 100);
+
+    for (tmp = input; *tmp; tmp++)
+        if (!strchr (SUPERSAFE, *tmp))
+            break;
+    if (!*tmp)
+    {
+        s_cat (&t, color);
+        s_cat (&t, input);
+        s_cat (&t, COLNONE);
+        return t.txt;
+    }
+    
+    s_cat (&t, color);
+    for (tmp = input; *tmp; tmp++)
+    {
+        if (*tmp == '\n' && !tmp[1])
+            ;
+        else if (*tmp & 0xe0 || (*tmp == '\n' && allownl))
+            s_catc (&t, *tmp);
+        else if (*tmp != '\r' || !allownl || tmp[1] != '\n')
+        {
+            s_catf (&t, COLINVCHAR);
+            s_catc (&t, *tmp - 1 + 'A');
+            s_catf (&t, color);
+        }
+    }
+    s_cat (&t, COLNONE);
+    return t.txt;
 }
 
